import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression


class Model:
    """The parent NMEC model class that handles the shared methods between all
        the child classes as well as handles the data processing to create
        models on."""

    def __init__(self, df=None, dependent_col="load", temperature_col="temp",
                 dependent_df=None, temperature_df=None, occupancy_col=None, 
                 additional_vars_df=None, model_name="TOWT"):
        """
        Initialize object by ensuring dataframe is fit for modeling and noting
        model specs.
        Parameters
        ----------
        df : Pandas DataFrame, optional
            A preprocessed dataframe that includes a timestamp column,
            dependent variable column, and other regressor variables. No other
            variables should be included. The default is None.
        dependent_col : str, optional
            Needed to mark what the dependent variable will be in the pre-
            processed dataframe. Not used if a preprocessed dataframe is not
            supplied. The default is "load".
        temperature_col : str, optional
            Needed to mark what the temperature variable is in the preprocessed
            dataframe, if a temperature column exists. If one is not needed for
            modeling, input None. Not used if a preprocessed dataframe is not
            supplied. The default is "temp".
        dependent_df : Pandas DataFrame, optional
            Dataframe containing a timestamp column and a dependent variable
            column. If wanting to train a model and a preprocessed dataframe
            is not supplied, this is needed. The default is None.
        temperature_df : Pandas DataFrame, optional
            Dataframe containing a timestamp column and a temperature column.
            If the desired model is assuming the dependent variable is
            dependent on temperature, this dataframe is needed. The default is
            None.
        additional_vars_df : Pandas DataFrame, optional
            Dataframe containing a timestamp column and other regressor
            variables. This is merged with the dependent_df dataframe.
            The default is None.
        model_name : str, optional
            Specific model name, which will be used in child subclasses.
            TOWT options are: TOWT, TOWT
            ChangePoint options are: SLR, 3PC, 3PH, 4P, 5P
            HDD_CDD options are: HD, CD, HD-CD
            default is towt.
            
        Raises
        ------
        KeyError
            DESCRIPTION.
        ValueError
            DESCRIPTION.
        Returns
        -------
        None.
        """
        if model_name is not None:
            self.model_name = model_name.upper()
        else:
            self.model_name = model_name

        self.interval_tiers = {'Monthly': 1, 'Daily': 2,
                               'Hourly': 3, '15 Minute': 4}
        self.model_interval = None
            
        if df is not None:
            if isinstance(df, pd.DataFrame):
                self.df = df
                self.timestamp_col = self.__check_timestamp_col(df)
                self.min_interval = self.infer_interval(self.df,
                                                        self.timestamp_col)
                self.dependent_col = dependent_col
                self.temperature_col = temperature_col
                self.occupancy_col = occupancy_col
                not_x_cols = [self.timestamp_col,
                              self.dependent_col,
                              self.temperature_col,
                              self.occupancy_col]
                # x_train_cols does not include temp column because it can make
                # a variable number of extra columns depending on modeling algo
                self.x_train_cols = list(self.df.drop(
                    columns=[x for x in not_x_cols if x is not None]).columns)
                try:
                    self.df[[dependent_col]]
                except KeyError:
                    raise KeyError(
                        "Supplied dependent variable column name is " +
                        dependent_col + " but it was not found in the supplied"
                        " dataframe.")
            else:
                raise ValueError("Supplied dataset is not a pandas dataframe.")
            if dependent_df is not None:
                raise ValueError("Supplied dataset is not a pandas dataframe.")


        elif dependent_df is not None:
            if isinstance(dependent_df, pd.DataFrame):
                self.temperature_col = None
                self.df = self.aggregate(dependent_df=dependent_df,
                                         temperature_df=temperature_df,
                                         additional_vars_df=additional_vars_df)
                self.occupancy_col = occupancy_col
                not_x_cols = [self.timestamp_col,
                              self.dependent_col,
                              self.temperature_col,
                              self.occupancy_col]
                self.x_train_cols = list(self.df.drop(
                    columns=[x for x in not_x_cols if x is not None]).columns)
                self.df.sort_values(by=self.timestamp_col)
            else:
                raise ValueError("Supplied dependent variable dataset is not a"
                                 "pandas DataFrame.")
        else:
            print("No data was provided for training.")
            self.df = None
            self.timestamp_col = None
            self.min_interval = None
            self.dependent_col = None
            self.temperature_col = None
            self.occupancy_col = None
            self.x_train_cols = None

        self.occ_model = None
        self.unocc_model = None

        if self.x_train_cols is not None:
            self.group_method = dict(
                zip([self.dependent_col, self.temperature_col,
                     self.occupancy_col] + self.x_train_cols,
                    ["sum", "mean", "mean"] + ["sum" for x in
                                               range(len(self.x_train_cols))]))
            self.group_method.pop(None, None)
            


    def aggregate(self, dependent_df=None, temperature_df=None,
                  additional_vars_df=None):
        """
        Aggregate dataframes containing the different possible variables for
        regression.
        Parameters
        ----------
        dependent_df : Pandas DataFrame, optional
            Dataframe consisting of two columns: a timestamp column and the
            dependent variable column. The default is None.
        temperature_df : Pandas DataFrame, optional
            Dataframe consisting of two columns: a timestamp column and a
            temperature column. The default is None.
        additional_vars_df : Pandas DataFrame, optional
            Dataframe consisting of at least two columns: a timestamp column
            and any amount of other variables to be included in the regression.
            The default is None.
        Returns
        -------
        None.
        """



        max_time_col = self.__check_timestamp_col(dependent_df)
        min_interval = self.infer_interval(dependent_df,
                                           max_time_col)
        main_df = dependent_df.copy()
        self.dependent_col = (dependent_df.drop(
            columns=[max_time_col]).columns[0])
        if temperature_df is not None:
            temperature_df_time_col = self.__check_timestamp_col(
                temperature_df)
            self.temperature_col = (temperature_df.drop(
                columns=[temperature_df_time_col]).columns[0])
            temperature_df_interval = self.infer_interval(
                temperature_df, temperature_df_time_col)
            main_df = main_df.merge(temperature_df, how="outer",
                                    left_on=max_time_col,
                                    right_on=temperature_df_time_col)
            if self.interval_tiers[min_interval] > self.interval_tiers[temperature_df_interval]:
                min_interval = temperature_df_interval
                max_time_col = temperature_df_time_col

        if additional_vars_df is not None:
            additional_vars_df_time_col = self.__check_timestamp_col(
                additional_vars_df)
            additional_vars_df_interval = self.infer_interval(
                additional_vars_df, additional_vars_df_time_col)
            main_df = main_df.merge(additional_vars_df, how="outer",
                                    left_on=max_time_col,
                                    right_on=additional_vars_df_time_col)
            if self.interval_tiers[min_interval] > self.interval_tiers[additional_vars_df_interval]:
                min_interval = additional_vars_df_interval
                max_time_col = additional_vars_df_time_col

        self.min_interval = min_interval
        self.timestamp_col = max_time_col
        main_df.sort_values(by=max_time_col, inplace=True)
        main_df.reset_index(drop=True, inplace=True)
        return main_df

    @staticmethod
    def __check_timestamp_col(df):
        """
        Function to identify which column is the timestamp column in a
        dataframe.
        Parameters
        ----------
        df : Pandas DataFrame
            Dataframe with a datetime column.
        Raises
        ------
        TypeError
            No timestamp column could be found in the dataframe.
        Returns
        -------
        list
            List of size 1 containing a string name of the datetime column.
        """

        timestamp_cols = []
        for i in df.columns:
            if ((pd.core.dtypes.common.is_datetime_or_timedelta_dtype(df[i]))
                    | (pd.api.types.is_datetime64_ns_dtype(df[i]))):
                timestamp_cols.append(i)
        if len(timestamp_cols) == 0:
            raise TypeError(
                'Could not find any datetime/timestamp columns')
        elif len(timestamp_cols) == 1:
            return(timestamp_cols[0])
            return(timestamp_cols)
        elif len(timestamp_cols) > 1:
            print('Multiple timestamp/datetime columns were found. '
                  'Assuming the first one is for the dataset.')
            timestamp_cols = timestamp_cols[0]
            return(timestamp_cols)

    def train_test_split(self, df, train_ratio=0.5, train_dates=()):
        pass

    def find_occ_unocc(self, hd_base=50, cd_base=65, time_interval="Monthly",
                       occ_threshold=0.65):
        """
        Determine occupancy schedule based off usage.
        Parameters
        ----------
        df : Pandas DataFrame
            Pandas dataframe consisting of at a minimum a timestamp,
            temperature, and load column labelled time, temp, and load.
        occ_threshold : float, optional
            DESCRIPTION. The default is 0.65.
        Returns
        -------
        None.
        """

        if time_interval == "Monthly":
            self.df['occ'] = 1
            return(self.df)
        if occ_threshold == 1:
            self.df['occ'] = 1
            return(self.df)

        orig_copy = self.df.copy()

        self.df['t50'] = self.df[self.temperature_col] - hd_base
        self.df.loc[self.df['t50'] > 0, 't50'] = 0
        self.df['t65'] = self.df[self.temperature_col] - cd_base
        self.df.loc[self.df['t65'] < 0, 't65'] = 0
        df_regress = self.df[[self.timestamp_col,
                              self.dependent_col, 't50', 't65']].dropna()
        x_regress = df_regress.loc[:, ['t50', 't65']]
        y_regress = df_regress[self.dependent_col]
        occ_test_model = LinearRegression().fit(x_regress, y_regress)
        df_regress['predict'] = occ_test_model.predict(x_regress)
        df_regress['residual'] = df_regress[self.dependent_col] - \
            df_regress['predict']
        df_regress['residual_count'] = df_regress['residual'] > 0
        df_regress['count'] = 1

        if time_interval == "Daily":
            df_regress['tow'] = df_regress[self.timestamp_col].dt.dayofweek
        elif time_interval == "Hourly":
            df_regress['tow'] = df_regress[self.timestamp_col].dt.dayofweek * \
                24 + df_regress[self.timestamp_col].dt.hour

        threshold_count = df_regress[
            ['tow', 'residual_count',
             'count']].groupby('tow').sum().reset_index()
        threshold_count['occ'] = (threshold_count['residual_count'] /
                                  threshold_count['count'] >
                                  occ_threshold).replace({True: 1, False: 0})
        df_regress = df_regress.merge(
            threshold_count[['tow', 'occ']], on="tow", how="outer")
        self.df_regress = df_regress
        self.orig_copy = orig_copy
        out = orig_copy.merge(
            df_regress[[self.timestamp_col, 'occ']], on=self.timestamp_col,
            how='outer').sort_values(by=self.timestamp_col)

        return(out)

    def format_results(self, prediction_df):
        """
        Parameters
        ----------
        prediction_df : TYPE
            DESCRIPTION.
        Returns
        -------
        None.
        """
        pass

    def train(self):
        pass

    def predict(self):
        pass

    @staticmethod
    def infer_interval(df, time_colname='time'):
        """
        Infers the time series interval in a supplied pandas dataframe.
        Parameters
        ----------
        df : pandas DataFrame
            Pandas DataFrame with at a minimum a timestamp column.
        time_colname : str or list of size 1 containing a str
            Value that provides what the timestamp column name is. The default
            is ['time'].
        Returns
        -------
        String that is Hourly, Daily, or Monthly.
        """

        # Get median number of minutes between observations
        minute_interval = df[time_colname].diff().apply(
            lambda x: x/np.timedelta64(1,
                                       'm')).fillna(0).astype('int64').median()

        if (minute_interval > 60/2) & (minute_interval < 60*2):
            time_interval = "Hourly"
            return(time_interval)
        elif(minute_interval > 10) & (minute_interval < 20):
            time_interval = "15 Minute"
            return(time_interval)
        elif (minute_interval > 1440/2) & (minute_interval < 1440*2):
            time_interval = "Daily"
            return(time_interval)
        elif (minute_interval > 40320/2) & (minute_interval < 40320*1.5):
            time_interval = "Monthly"
            return(time_interval)
        else:
            raise ValueError(
                "Could not infer time interval based off 'time' column in"
                " dataframe.")

    def group_interval(self, df=None, time_interval=None, group_method=None,
                       time_col="time"):
        """
        Parameters
        ----------
        df : TYPE, optional
            DESCRIPTION. The default is None.
        time_interval : TYPE, optional
            DESCRIPTION. The default is None.
        group_method : TYPE, optional
            DESCRIPTION. The default is None.
        time_col : TYPE, optional
            DESCRIPTION. The default is "time".
        Raises
        ------
        ValueError
            DESCRIPTION.
        Returns
        -------
        None.
        """

        # Checks
        if df is None:
            if self.df is not None:
                df = self.df.copy()
            else:
                raise ValueError("No dataframe was provided.")
        if time_interval is None:
            if self.min_interval is not None:
                time_interval = self.min_interval
            else:
                raise ValueError("No time interval was provided.")
        if group_method is None:
            if self.group_method is not None:
                group_method = self.group_method
            else:
                raise ValueError("No grouping method was provided for"
                                 " variables.")
        df['year'] = df[time_col].dt.year
        df['month'] = df[time_col].dt.month
        df['day'] = df[time_col].dt.day
        df['hour'] = df[time_col].dt.hour

        if time_interval == 'Hourly':
            temp_time_cols = ['year', 'month', 'day', 'hour']
        elif time_interval == 'Daily':
            temp_time_cols = ['year', 'month', 'day']
        elif time_interval == 'Monthly':
            temp_time_cols = ['year', 'month']

        main = df.groupby(temp_time_cols).agg(group_method).reset_index()
        main[time_col] = pd.to_datetime(main[temp_time_cols])
        main.drop(columns=temp_time_cols, inplace=True)

        return(main)


'''
    @staticmethod
    def r2(df):
        rss = ((df['load'] - df['predict']) ** 2).sum()
        y_bar = df['load'].mean()
        tss = ((df['load'] - y_bar) ** 2).sum()
        r2 = 1 - (rss / tss)

        return np.round(r2, 4)

    def adj_r2(self, df):
        r_squared = self.r2(df)
        rt = ((1 - r_squared) * (self.records - 1)) / \
            (self.records - self.predicted_columns - 1)  # right term
        adj_r_squared = 1 - rt

        return np.round(adj_r_squared, 4)

    def cvrmse(self, df) -> np:
        """
        Coefficient of Variation of the Root Mean Square Error (Percentage)
        :param df:  pandas dataframe consisting of the data the model was built
                    off of with the predicted values
        """
        rmse = (((df['load'] - df['predict']) ** 2).sum() /
                (self.records - self.predicted_columns)) ** (1 / 2)
        cvrmse = rmse / df['load'].mean()

        return np.round(cvrmse, 4)

    @staticmethod
    def nmbe(df):
        nmbe = (df['load'] - df['predict']).sum() / df['load'].sum()

        return np.round(nmbe, 4)

    def metrics(self, df):
        return {'R2': self.r2(df),
                'adj R2': self.adj_r2(df),
                'CVRMSE': self.cvrmse(df),
                'NMBE': self.nmbe(df)}
'''



from .Model import Model
import pandas as pd
from sklearn.linear_model import LinearRegression


class TOWTModel(Model):

    def __init__(self, df=None, dependent_col="load", temperature_col="temp",
                 occupancy_col=None, dependent_df=None, temperature_df=None,
                 additional_vars_df=None, occ_threshold=0.65,
                 knot_temps=[40, 55, 65, 80, 90], model_name="TOWT"):
        """
        Initialize object by ensuring data provided is fit for modeling and
        noting model specs.
        Parameters
        ----------
        df : Pandas DataFrame, optional
            A preprocessed dataframe that includes a timestamp column,
            dependent variable column, and other regressor variables. No other
            variables should be included. The default is None.
        dependent_col : str, optional
            Needed to mark what the dependent variable will be in the pre-
            processed dataframe. Not used if a preprocessed dataframe is not
            supplied. The default is "load".
        temperature_col : str, optional
            Needed to mark what the temperature variable is in the preprocessed
            dataframe, if a temperature column exists. If one is not needed for
            modeling, input None. Not used if a preprocessed dataframe is not
            supplied. The default is "temp".
        occupancy_col : str, optional
            Needed to mark what the binary occupied variable is in the
            preprocessed dataframe or additional variables dataframe, if a
            occupancy column exists. If one is not supplied, an occupancy
            schedule can be estimated based off the load. If separate
            occupied/unoccupied models are not wanted, change the occ_threshold
            to 1. The default is None.
        dependent_df : Pandas DataFrame, optional
            Dataframe containing a timestamp column and a dependent variable
            column. If wanting to train a model and a preprocessed dataframe
            is not supplied, this is needed. The default is None.
        temperature_df : Pandas DataFrame, optional
            Dataframe containing a timestamp column and a temperature column.
            If the desired model is assuming the dependent variable is
            dependent on temperature, this dataframe is needed. The default is
            None.
        additional_vars_df : Pandas DataFrame, optional
            Dataframe containing a timestamp column and other regressor
            variables. This is merged with the dependent_df dataframe.
            The default is None.
        occ_threshold : float, optional
            Percentage of time assumed to be occupied. The default is 0.65.
        knot_temps : list, optional
            List of knot temperatures for TOWT models. The default is
            [40, 55, 65, 80, 90].
        model_name : str, optional
            Specific model name.
            Options are: TOW, TOWT
            Upper or lower case is accepted. Default is TOWT.
            
        Returns
        -------
        None.
        """
        super().__init__(df=df,
                         dependent_col=dependent_col,
                         temperature_col=temperature_col,
                         dependent_df=dependent_df,
                         temperature_df=temperature_df,
                         additional_vars_df=additional_vars_df,
                         occupancy_col=occupancy_col,
                         model_name=model_name)
        self.temp_train_cols = []
        self.tow_train_cols = []
        self.knot_temps = knot_temps
        self.occ_threshold = occ_threshold
        
        if self.model_name not in ["TOWT", "TOW"]:
            raise ValueError("Unknonwn model type: " + self.model_name + ". "
                             "Please specify model_name as TOW or TOWT.")

    def train(self, interval=None, knot_temps=None):
        """
        Training method for TOWT and TOW model.
        Parameters
        ----------
        model : str, optional
            String that to run either time of week and temperature model or
            just time of week. The default is "TOWT".
        knot_temps : list, optional
            List of knot_temps at least of length 2. Will rewrite class
            knot_temps attribute. The default is None.
        Raises
        ------
        ValueError
            If inferred time interval is monthly, the error will raise.
        Returns
        -------
        Pandas DataFrame
            The main dataframe with all the temp knot and time of week
            variables used for regression attached to it.
        """
        # Checks
        if (knot_temps != self.knot_temps) & (knot_temps is not None):
            self.knot_temps = knot_temps

        if interval is None:
            interval = self.min_interval
        
        min_available_interval_num = self.interval_tiers[self.min_interval]
        desired_interval_num = self.interval_tiers[interval]

        if min_available_interval_num > desired_interval_num:
            self.df = self.group_interval(time_interval=interval,
                                          time_col=self.timestamp_col)
         
        if min_available_interval_num < desired_interval_num:
            raise ValueError(
                'Desired interval: ' + interval + 
                ' is unavailable. Highest granularity is ' + 
                self.min_interval + '.')

        # Check if time_interval is monthly
        if interval == "Monthly":
            raise ValueError(
                'Inferred time interval is monthly. The TOWT'
                ' algorithm expects hourly or daily data.')
        
        self.model_interval = interval

        # Define knot temp vars and tow vars
        used_knot_temps = pd.DataFrame()
        temp_vars = pd.DataFrame()
        self.x_train_cols = list(set(self.x_train_cols)
                                 - set(self.temp_train_cols
                                       + self.tow_train_cols))
        if self.model_name == "TOWT":
            temp_vars, used_knot_temps = self.define_temp_vars(
                self.df, self.knot_temps)
            self.temp_train_cols = list(temp_vars.columns)
            self.x_train_cols = self.x_train_cols + self.temp_train_cols
        tow_vars = self.define_tow_vars(self.df, self.model_interval)
        self.tow_train_cols = list(tow_vars.columns)
        self.x_train_cols = self.x_train_cols + self.tow_train_cols

        # Assemble regression datasets
        if self.occupancy_col is None:
            self.df = self.find_occ_unocc(time_interval=self.model_interval)
            self.occupancy_col = 'occ'

        main = pd.concat([self.df, temp_vars, tow_vars], axis=1)
        train = main.dropna().copy()

        # Create models and predict
        # Occupied model
        occ_x_train = train.loc[train[self.occupancy_col]
                                == 1, self.x_train_cols]
        occ_y_train = train.loc[train[self.occupancy_col]
                                == 1, self.dependent_col]
        occ_model = LinearRegression().fit(occ_x_train, occ_y_train)
        self.occ_model = occ_model

        # Unoccupied model
        if train[self.occupancy_col].nunique() > 1:
            unocc_x_train = train.loc[train[self.occupancy_col]
                                      == 0, self.x_train_cols]
            unocc_y_train = train.loc[train[self.occupancy_col]
                                      == 0, self.dependent_col]
            unocc_model = LinearRegression().fit(unocc_x_train, unocc_y_train)
            self.unocc_model = unocc_model

        '''
        occ_model, occ_train = model_and_predict(train=train, occupied=True)
        # Unoccupied model
        unocc_model = None
        unocc_train = pd.DataFrame()
        if train['occ'].nunique() > 1:
            unocc_model, unocc_train = model_and_predict(
                train=train, occupied=False)
        # Merge dataset back together
        train_estimate = pd.concat([occ_train, unocc_train])
        train_estimate = pd.concat([main, train_estimate['predict']], axis=1)
        # Get metrics
        metrics = {'R2': R2(train_estimate),
                   'adj R2': adjR2(train_estimate),
                   'CVRMSE': cvrmse(train_estimate),
                   'NMBE': nmbe(train_estimate)}
        # Create dictionary for model/function outputs
        output = {"occupied model": occ_model,
                  "unoccupied model": unocc_model,
                  "data": train_estimate,
                  'knot_temps': (knot_temps, used_knot_temps),
                  'metrics': metrics}
        return(output)
        '''
        return(main)

    def predict(self, df, occ_model=None, unocc_model=None):
        """
        Predict method for TOWT and TOW models.
        Parameters
        ----------
        df : Pandas DataFrame
            Dataframe consisting of at least time['time'] and occupancy['occ'].
            Temp is needed for TOWT.
        occ_model : sklearn LinearRegression object, optional
            Regression object for occipied model. This is required. If None,
            will assume it is an object attribute.
        unocc_model : sklearn LinearRegression object, optional
            Regression object for unoccipied model. The default is None.
            If None, will assume it is an object attribute. If object attribute
            is also None, will assume there is no unoccupied model to predict
            on.

        Returns
        -------
        The main pandas dataframe with the predictions attached to it.
        """

        # Check model objects
        if (occ_model != self.occ_model) & (occ_model is not None):
            self.occ_model = occ_model
        if (unocc_model != self.unocc_model) & (unocc_model is not None):
            self.unocc_model = unocc_model
        time_interval = self.infer_interval(df)

        # Check if time_interval is monthly
        if time_interval == "Monthly":
            raise ValueError(
                'Inferred time interval is monthly. The TOWT'
                ' algorithm expects hourly or daily data.')

        # Define knot temp vars and tow vars
        used_knot_temps = pd.DataFrame()
        temp_vars = pd.DataFrame()
        if self.model_name == "TOWT":
            temp_vars, used_knot_temps = self.define_temp_vars(
                df, self.knot_temps)
        tow_vars = self.define_tow_vars(df, time_interval)

        main = pd.concat([df, temp_vars, tow_vars], axis=1)
        pred = main.dropna().copy()
        if self.model_name == "TOWT":
            pred.drop(columns=['time', 'temp'], inplace=True)
        else:
            pred.drop(columns=['time'], inplace=True)

        # Predict
        occ_x_pred = pred.loc[pred['occ'] == 1, pred.columns != 'load']
        occ_x_pred.drop(columns='occ', inplace=True)
        occ_x_pred['predict'] = occ_model.predict(occ_x_pred)
        if pred['occ'].nunique() > 1:
            unocc_x_pred = pred.loc[pred['occ'] == 0, pred.columns != 'load']
            unocc_x_pred.drop(columns='occ', inplace=True)
            unocc_x_pred['predict'] = unocc_model.predict(unocc_x_pred)

        pred = pd.concat([occ_x_pred, unocc_x_pred], axis=0)
        out = pd.concat([main, pred['predict']], axis=1)
        return(out)

    def temp0(self, temp, knot_temp):
        """
        Defining first knot temp values in TOWT model.
        Parameters
        ----------
        temp : float
            Temperature value from the 'temp' column in the main dataframe.
        knot_temp : float
            The lowest knot temp.
        Returns
        -------
        float
            The first knot temperature variable value for TOWT model.
        """

        if temp > knot_temp:
            return(knot_temp)
        else:
            return(temp)

    def tempi(self, temp, knot_tempi, knot_tempj):
        """
        Defining the middle knot temp values in TOWT model.
        Parameters
        ----------
        temp : float
            Temperature value from the 'temp' column in the main dataframe.
        knot_tempi : float
            A knot temperature that is not the lowest or greatest.
        knot_tempj : float
            A knot temperature that is one index lower than i
        Returns
        -------
        float
            A middle knot temperature variable value in the TOWT model.
        """

        if temp > knot_tempi:
            return(knot_tempi - knot_tempj)
        else:
            if temp > knot_tempj:
                return(temp - knot_tempj)
            else:
                return(0)

    def tempn(self, temp, knot_tempn):
        """
        Defining the last knot temp value in TOWT model.
        Parameters
        ----------
        temp : float
            Temperature value from the 'temp' column in the main dataframe.
        knot_tempn : float
            The highest knot temperature.
        Returns
        -------
        float
            The last knot temperature variable value.
        """
        if temp > knot_tempn:
            return(temp - knot_tempn)
        else:
            return(0)

    def define_temp_vars(self, df, knot_temps):
        """
        Define all knot temperature variable values given a dataframe with
        temperature and specific knot temperatures.
        Parameters
        ----------
        df : pandas dataframe
            The main pandas dataframe that will be used for analysis with a
            numeric temperature column.
        knot_temps : list
            List of specific knot temperatures.
        Returns
        -------
        towt_vars: Pandas DataFrame
            Dataframe consisting of the TOWT variables.
        knot_temps : TYPE
            Used knot temperatues to determine TOWT variables.
        """
        # Define min and max temperatures in dataset
        min_temp = df['temp'].min()
        max_temp = df['temp'].max()

        # Drop outside knot temperature bounds if redundant
        knot_temps = [x for x in knot_temps if x > min_temp]
        knot_temps = [x for x in knot_temps if x < max_temp]

        # Sort knot temperatures from least to greatest
        knot_temps.sort()

        # Create the temperature variables according to LBNL algo (Mathieu)
        temp_var_dict = {}
        for i in range(len(knot_temps)):
            idx = 'temp'+str(i)
            if i == 0:
                temp_var_dict[idx] = df['temp'].apply(
                    self.temp0, args=(knot_temps[i],)).values
            elif i == len(knot_temps)-1:
                temp_var_dict[idx] = df['temp'].apply(
                    self.tempi, args=(knot_temps[i], knot_temps[i-1])).values
                temp_var_dict['temp'+str(i+1)] = df['temp'].apply(self.tempn,
                                                                  args=(knot_temps[i],)).values
            else:
                temp_var_dict[idx] = df['temp'].apply(
                    self.tempi, args=(knot_temps[i], knot_temps[i-1])).values

        return pd.DataFrame(temp_var_dict), knot_temps

    @staticmethod
    def define_tow_vars(df, time_interval):
        """
        The function creates time of week indicator variables. 168 hours of
        week or 7 days of week.
        Parameters
        ----------
        df : Pandas DataFrame
            Dataframe consisting of at a minimum a timestamp column labelled
            "time".
        time_interval : str
            DESCRIPTION.
        Returns
        -------
        Pandas DataFrame
            Dataframe containing indicator variables for time of week.
        """
        if time_interval.upper() == 'DAILY':
            dow = df['time'].dt.dayofweek
            tow_vars = pd.get_dummies(dow, prefix="tow")
            return(tow_vars)

        elif time_interval.upper() == 'HOURLY':
            dow = df['time'].dt.dayofweek
            hour = df['time'].dt.hour
            tow = dow*24 + hour
            tow_vars = pd.get_dummies(tow, prefix="tow")
            return(tow_vars)

        else:
            raise ValueError(
                "Time interval supplied expected 'daily' or 'hourly'. "
                "Neither of those were given.")



import os
import numpy as np
import pandas as pd
import seaborn as sns
from .ChangePointModel import ChangePointModel

class NMEC:
    """
    An implementation of peer-reviewed energy data analysis algorithms in Python for site-specific M&V.

    data_path: Str, optional Path to a file containing all data. The default is None.
    data_intervals: List, optional List of the intervals you want to create. If nothing is passed then possible intervals will be inferred. The default is [].
    eload_path: Str, optional Path to a file containing eload data. The default is None.
    temp_path: Str, optional Path to a file containing temp data. The default is None.
    additional_path: Str, optional Path to a file containing additional data. The default is None.
    """
    df = None

    def __init__(self, data_path=None, data_intervals=None, eload_path=None, 
                 temp_path=None, additional_path=None):
        if data_intervals is None:
            data_intervals = []
        self.data_path = data_path
        self.eload_path = eload_path
        self.temp_path = temp_path
        self.additional_path = additional_path
        self.data_intervals = data_intervals

    def load_data(self, path=""):
        """
        Loads data given the data paths


        Parameters
        ----------
        path : Str
            Data path that you want to load from. The default is "".

        Returns
        -------
        df: An unformatted dataframe, list, dict, or array
        """
        pass

    def format_df(self):
        """
        Formats the data for usage in Models
        Returns
        -------
        df: A formatted pandas df
        """
        pass

    def model_with_changepoint(self, df: pd, cp_type="SLR"):
        my_model = ChangePointModel(df, cp_type)
        
        if cp_type.upper() in ["3PH", "3PC"]:
            df = my_model.df
            train_estimate = my_model.train(df.dropna().copy())

            return {"occupied model": my_model.p, "unoccupied model": None,
                    "data": train_estimate
                    # 'metrics': my_model.metrics(train_estimate)
                    }

        elif cp_type.upper() in ["SLR", "2P"]:
            
            if 'occ' in my_model.df.columns:
                df_occ = my_model.df.loc[df['occ'] == 1]
                df_unocc = my_model.df.loc[df['occ'] == 0]
            else:
                df_occ = my_model.df
                df_unocc = None
                
            train_estimate = my_model.train_SLR(df_occ.dropna().copy())
            
            if df_unocc is not None:
                train_estimate = my_model.train_SLR(
                    df_unocc.dropna().copy(), train_estimate)

            if train_estimate is not None:
                return {"occupied model": my_model.lm, "unoccupied model": my_model.lm_unocc,
                        "data": train_estimate
                        # 'metrics': my_model.metrics(train_estimate)
                        }
            else:
                raise ValueError(
                    'No Training Estimate, check if Type is SLR, 2P, 3PH, or 3PC')
  

    def example(self, model_type=None):
        print("modeling with type", model_type)

        df = pd.read_csv(
            f'{os.getcwd()}/tests/data/example.csv', parse_dates=[1])

        # Clean up Load
        df['load'] = df['load'].astype(float)

        # Clean up time
        df['time'] = pd.to_datetime(df['time'], errors='coerce', utc=True)
        df['time'] = df['time'].dt.tz_convert('Etc/GMT+7')

        # Print results
        result = self.model_with_changepoint(df, cp_type = model_type)
        
        return result
    
    def plot_results(self, df):

        df_long = pd.melt(df, id_vars=['time', 'temp', 'occ'], value_name="eload", var_name="Measurement")
        df_long = df_long.replace({'load': 'Actual', 'predict': 'Prediction'}) 
        # df_long['occ'] = train_long['occ'].replace({0: 'Unoccupied', 1: 'Occupied'})    

        sns_plot = sns.relplot(data=df_long, x='temp', y='eload', hue="Measurement", s=40, linewidth=0)
        return sns_plot



from .Model import Model
import numpy as np
import pandas as pd
from scipy import optimize, stats


class ChangePointModel(Model):

    def __init__(self, df=None, dependent_col="load", temperature_col="temp",
                 dependent_df=None, temperature_df=None,
                 additional_vars_df=None, model_name="3PH"):
        """
        Initialize object by ensuring data provided is fit for modeling and
        noting model specs.
        Parameters
        ----------
        df : Pandas DataFrame, optional
            A preprocessed dataframe that includes a timestamp column,
            dependent variable column, and other regressor variables. No other
            variables should be included. The default is None.
        dependent_col : str, optional
            Needed to mark what the dependent variable will be in the pre-
            processed dataframe. Not used if a preprocessed dataframe is not
            supplied. The default is "load".
        temperature_col : str, optional
            Needed to mark what the temperature variable is in the preprocessed
            dataframe, if a temperature column exists. If one is not needed for
            modeling, input None. Not used if a preprocessed dataframe is not
            supplied. The default is "temp".
        occupancy_col : str, optional
            Needed to mark what the binary occupied variable is in the
            preprocessed dataframe or additional variables dataframe, if a
            occupancy column exists. If one is not supplied, an occupancy
            schedule can be estimated based off the load. If separate
            occupied/unoccupied models are not wanted, change the occ_threshold
            to 1. The default is None.
        dependent_df : Pandas DataFrame, optional
            Dataframe containing a timestamp column and a dependent variable
            column. If wanting to train a model and a preprocessed dataframe
            is not supplied, this is needed. The default is None.
        temperature_df : Pandas DataFrame, optional
            Dataframe containing a timestamp column and a temperature column.
            If the desired model is assuming the dependent variable is
            dependent on temperature, this dataframe is needed. The default is
            None.
        additional_vars_df : Pandas DataFrame, optional
            Dataframe containing a timestamp column and other regressor
            variables. This is merged with the dependent_df dataframe.
            The default is None.
        model_name : str, optional
            Specific model name.
            Options are: SLR, 3PC, 3PH, 4P, 5P.
            Upper or lower case is accepted. Default is 3PH.

        Raises
        ------
        ValueError
            If specifed changepoint type is not recognized or not provided
        Returns
        -------
        None.
        """
        super().__init__(df=df,
                         dependent_col=dependent_col,
                         temperature_col=temperature_col,
                         dependent_df=dependent_df,
                         temperature_df=temperature_df,
                         additional_vars_df=additional_vars_df,
                         model_name=model_name)

        if self.model_name not in ["SLR", "3PH", "3PC", "4PH", "4PC", "5P"]:
            raise ValueError("Unknonwn model type: " + self.model_name + ". "
                             "Please specify model_name as"
                             " 'SLR', '3PH', or '3PC', '4PH, '4PC' or '5P'.")
            
        if self.model_name == "SLR":
            self.model = self.SLR
            self.feature_names_in_ = ["slp", "y_int"]     
            
        if self.model_name == "3PH":
            self.model = self.piecewise_3PH
            self.feature_names_in_ = ["hcp", "base", "hsl"]
            
        if self.model_name == "3PC":
            self.model = self.piecewise_3PC
            self.feature_names_in_ = ["ccp", "base", "csl"]
        
        if self.model_name == "4PH":
            self.model = self.piecewise_4PH
            self.feature_names_in_ = ["hcp", "base", "hsl1", "hsl2"]
            
        if self.model_name == "4PC":
            self.model = self.piecewise_4PC
            self.feature_names_in_ = ["ccp", "base", "csl1", "csl2"]
            
        # self.occ_model = self.model
        # self.unocc_model = None
            
    @staticmethod
    def SLR(x, slp, y_int):
        """ 3-parameter heating functional form
        
        Parameters
        ----------
        x: array, temperatures
        slp: float, slope
        y_int: float, y-intercept
        
        Returns
        -------
        numpy piecewise function
        
        """
        conds = [x]
        
        funcs = [lambda x: slp * x + y_int]
        
        return np.piecewise(x, conds, funcs)

        
    @staticmethod
    def piecewise_3PH(x, hcp, base, hsl):
        """ 3-parameter heating functional form
        
        Parameters
        ----------
        x: array, temperatures
        hcp: float, heating change point
        base: float, non-heating constant
        hsl: float, heating slope
        
        Returns
        -------
        numpy piecewise function
        
        """
        conds = [x < hcp, x >= hcp]
        
        funcs = [lambda x: hsl * (x-hcp) + base,
                 lambda x: base]
        
        return np.piecewise(x, conds, funcs)

    @staticmethod
    def piecewise_3PC(x, ccp, base, csl):
        """ 3-parameter cooling functional form
        
        Parameters
        ----------
        x: array, temperatures
        ccp: float, cooling change point
        csl: float, cooling slope
        base: float, non-heating constant

        Returns
        -------
        numpy piecewise function

        """
        conds = [x <= ccp, x > ccp]
        
        funcs = [lambda x: base,
                 lambda x: csl * (x-ccp) + base]
        
        return np.piecewise(x, conds, funcs)
    
    @staticmethod
    def piecewise_4PH(x, hcp, base, hsl1, hsl2):
        """ 4-parameter heating functional form
        
        Parameters
        ----------
        x: array, temperatures
        hcp: float, change point
        base: float, constant
        hsl1: float, heating slope 1
        hsl2: float, heating slope 2
        
        Returns
        -------
        numpy piecewise function
        
        """
        conds = [x < hcp, x >= hcp]
        
        funcs = [lambda x: hsl1 * (x-hcp) + base,
                 lambda x: hsl2 * (hcp-x) + base]
        
        return np.piecewise(x, conds, funcs)
    
    @staticmethod
    def piecewise_4PC(x, ccp, base, csl1, csl2):
        """ 4-parameter cooling functional form
        
        Parameters
        ----------
        x: array, temperatures
        ccp: float, change point
        base: float, constant
        csl1: float, cooling slope 1
        csl2: float, cooling slope 2
        
        Returns
        -------
        numpy piecewise function
        
        """
        conds = [x < ccp, x >= ccp]
        
        funcs = [lambda x: csl1 * (x-ccp) + base,
                 lambda x: csl2 * (ccp-x) + base]
        
        return np.piecewise(x, conds, funcs)
    
    

    def get_bounds(self, temperature):
        """
        Get temperature bounds for piecewise linear parameters

        Parameters
        ----------
        temperature : array, temperatures

        Returns
        -------
        bounds: dictinary of values

        """
        hcp_bound_percentile = 45
        ccp_bound_percentile = 55

        percentiles = hcp_bound_percentile, ccp_bound_percentile
        # heating change point, cooling change point
        hcp, ccp = np.percentile(temperature, percentiles)

        hcp_min = hcp  # Heating change-point minimum
        hcp_max = ccp  # Heating change-point maximum
        ccp_min = hcp  # Cooling change-point minimum
        ccp_max = ccp  # Cooling change-point minimum
        base_min = 0  # Baseload minimum
        base_max = np.inf  # Baseload maximum
        hsl_min = -np.inf  # Heating slope minimum
        hsl_max = 0  # Heating slope maximum
        csl_min = 0  # Cooling slope minimum
        csl_max = np.inf  # Cooling slope maximum
        
        if self.model_name == "SLR":
            training_bounds = ([hsl_min, base_min],
                               [csl_max, base_max])
            
        elif self.model_name == "3PH":
            training_bounds = ([hcp_min, base_min, hsl_min], 
                               [hcp_max, base_max, hsl_max])
            
        elif self.model_name == "3PC":
            training_bounds = ([ccp_min, base_min, csl_min], 
                               [ccp_max, base_max, csl_max])
            
        elif self.model_name == "4PH":
            training_bounds = ([hcp_min, base_min, hsl_min, hsl_min], 
                               [hcp_max, base_max, hsl_max, hsl_max])
            
        elif self.model_name == "4PC": 
            training_bounds = ([ccp_min, base_min, csl_min, csl_min], 
                               [ccp_max, base_max, csl_max, csl_max])
            
        else:
            raise ValueError(
                'Incorrect changepoint type specified: ' + self.model_name + '.')
        return training_bounds
    
    
    def fit_model(self, x, y):
        """
        Fit a change point model to the data.

        Parameters
        ----------
        x : numpy array
            Temperature
        y : numpy array
            Load

        Returns
        -------
        y_fit : numpy array
            Fitted load values

        """
        x = np.array(x)
        y = np.array(y)
        model_bounds = self.get_bounds(temperature = x)
        
        p, e = optimize.curve_fit(f = self.model, 
                                  xdata = x, 
                                  ydata = y, 
                                  bounds = model_bounds)
        y_fit = self.model(x, *p)
        return p, y_fit
    

    def train(self, interval=None):
        """
        Training method for change point models (SLR, 3PC, 3PH, 4P, 5P)

        Parameters
        ----------
        interval: str, optional
            Desired interval to aggregate data to for training. 
            If None is provided then the finest granulairty possible is used.

        Returns
        -------
        Pandas Dataframe 
            Origianl training data with fitted values added
        """
        if interval is None:
            interval = self.min_interval
        
        min_available_interval_num = self.interval_tiers[self.min_interval]
        desired_interval_num = self.interval_tiers[interval]

        if min_available_interval_num > desired_interval_num:
            self.df = self.group_interval(time_interval=interval,
                                          time_col=self.timestamp_col)
         
        if min_available_interval_num < desired_interval_num:
            raise ValueError(
                'Desired interval: ' + interval + 
                ' is unavailable. Finest granularity interval is ' + 
                self.min_interval + '.')
        self.model_interval = interval
        
        train = self.df.dropna().copy()
        
        x_train = train[self.temperature_col]
        y_train = train[self.dependent_col]
        
        self.coef_ , y_fit = self.fit_model(x_train, y_train)
        
        train_estimate = pd.DataFrame({'time': train[self.timestamp_col],
                                   'temp': train[self.temperature_col],
                                   'load': train[self.dependent_col],
                                   'y_fit': y_fit}) 
        return train_estimate


    def predict(self, df=None, model=None):
        
        if df is None:
            raise ValueError(
                'No prediction df specified. Please pass a df to predict on.')
        if self.model is None:
            raise ValueError(
                'No trained model. Please train a model.')
        
        # predict_ts_col = self.__check_timestamp_col(df)

        predict_interval = self.infer_interval(df)
        
        min_predict_interval_num = self.interval_tiers[predict_interval]
        model_interval_num = self.interval_tiers[self.model_interval]

        if min_predict_interval_num > model_interval_num:
            df = self.group_interval(time_interval=self.model_interval,
                                          time_col=self.timestamp_col)
        
        if min_predict_interval_num < model_interval_num:
            raise ValueError(
                'Modeled interval: ' + self.model_interval + 
                ' does not match data interval: ' + predict_interval + '.')
        
        main = df.dropna().copy()
        x_pred = np.array(main[self.temperature_col])
        y_pred = self.model(x_pred, *self.coef_)
        
        predictions = pd.DataFrame({'time': main[self.timestamp_col],
                                   'temp': main[self.temperature_col],
                                   'y_pred': y_pred}) 
        
        return predictions
        



from setuptools import setup

setup(
    name='nmecpy',
    version='0.1.5',    
    description='An implementation of peer-reviewed energy data analysis algorithms in Python for site-specific M&V',
    url='https://github.com/kW-Labs/nmecpy',
    author='Maggie Jacoby, Deter Luu, Devan Johnson',
    author_email='dbardin@kw-engineering.com',
    license='MIT',
    packages=['nmecpy'],
    install_requires=['sklearn',
                      'pandas',
                      'numpy',
                      'scipy',
                      'seaborn', 
                      'warnings',            
                      ],

    classifiers=[
        
    ],
)


from .Model import Model
import pandas as pd
from sklearn.linear_model import LinearRegression


class HDDCDDModel(Model):

    def __init__(self, df=None):
        """
        The class assumes a dataframe with HD and CD values that are already
        calculated.

        Parameters
        ----------
        df : Pandas DataFrame, optional
            If training a model, a dataframe must be provided. If using only
            for predictions, a dataframe is not needed when initialized.

        Returns
        -------
        None.

        """
        super().__init__(df=df)

    def train(self, model="HD-CD"):
        """
        Model training method for heating degree and cooling degree models.

        Parameters
        ----------
        model : str, optional
            String value equaling HD-CD for a heating and cooling degree model.
            HD for just a heating degree model. CD for just a cooling degree
            model. The default is "HD-CD".

        Returns
        -------
        None.

        """

        time_interval = self.infer_interval(self.df)

        # Assemble regression datasets
        if ('occ' not in self.df.columns):
            self.df = self.find_occ_unocc(time_interval=time_interval)

        main = self.df
        train = self.df.dropna().copy()
        train.drop(columns=['time', 'temp'], inplace=True)

        # Create models and predict
        # Occupied model
        occ_x_train = train.loc[train['occ'] == 1, train.columns != 'load']
        occ_x_train.drop(columns='occ', inplace=True)
        occ_y_train = train.loc[train['occ'] == 1, 'load']
        occ_model = LinearRegression().fit(occ_x_train, occ_y_train)
        self.occ_model = occ_model

        # Unoccupied model
        if train['occ'].nunique() > 1:
            unocc_x_train = train.loc[train['occ']
                                      == 0, train.columns != 'load']
            unocc_x_train.drop(columns='occ', inplace=True)
            unocc_y_train = train.loc[train['occ'] == 0, 'load']
            unocc_model = LinearRegression().fit(unocc_x_train, unocc_y_train)
            self.unocc_model = unocc_model

        return(main)

    def predict(self, df, occ_model=None, unocc_model=None):
        """
        Predict method for heating degree and cooling degree methods.

        Parameters
        ----------
        df : Pandas DataFrame
            Dataframe with heating and/or cooling degree variables.
        occ_model : sklearn LinearRegression object, optional
            Regression object for occipied model. This is required. If None,
            will assume it is an object attribute.
        unocc_model : sklearn LinearRegression object, optional
            Regression object for unoccipied model. The default is None.
            If None, will assume it is an object attribute. If object attribute
            is also None, will assume there is no unoccupied model to predict
            on.

        Returns
        -------
        Predictions attached to main dataframe.

        """

        # Check model objects
        if (occ_model != self.occ_model) & (occ_model is not None):
            self.occ_model = occ_model
        if (unocc_model != self.unocc_model) & (unocc_model is not None):
            self.unocc_model = unocc_model

        main = df
        pred = main.dropna().copy()
        pred.drop(columns=['time', 'temp'], inplace=True)

        # Predict
        occ_x_pred = pred.loc[pred['occ'] == 1, pred.columns != 'load']
        occ_x_pred.drop(columns='occ', inplace=True)
        occ_x_pred['predict'] = occ_model.predict(occ_x_pred)
        if pred['occ'].nunique() > 1:
            unocc_x_pred = pred.loc[pred['occ'] == 0, pred.columns != 'load']
            unocc_x_pred.drop(columns='occ', inplace=True)
            unocc_x_pred['predict'] = unocc_model.predict(unocc_x_pred)

        pred = pd.concat([occ_x_pred, unocc_x_pred], axis=0)
        out = pd.concat([main, pred['predict']], axis=1)
        return(out)
